<?php
/**
 * This file contains the image class to use in CMS Contenido version 4.9.x and 4.10.x.
 *
 * @author Bodil Schwerer
 * @version 1.1
 *
 * changes
 * 1.1 - webp may be transparent
 *
 * @copyright Fjordsites <fjordsites.de>
 * @link http://fjordsites.de
 */

defined('CON_FRAMEWORK') || die('Illegal call: Missing framework initialization - request aborted.');

/**
 * cImage
 * Class to resize, crop and store cache images
 *
 */	
	class cImage {

/*                   *
 * SYSTEM PROPERTIES *
 *                   */
		
    /**
     * Contenido database resource
     *
     * @var resource
     */
		private $_db;
		
    /**
     * current idclient 
     *
     * @var int
     */
		
		private $_idclient = 0;
		
    /**
     * client configuration array
     *
     * @var array
     */ 
		private $cfgClient = array();
		
    /**
     * system configuration array
     *
     * @var array
     */ 
		private $cfg = array();
		
		
		private $cachePath = '';
		
		
		private $cacheFrontendPath = '';
		
	/**
	 * path to an cach dir. May help keeping images when regular cache gets empty.
	 *
	 * @var string
	 */
	 
		private $relativeCacheDir = '';
		
    /**
     * array to collect log information
     *
     * @var array
     */ 
		public $log = array();
		
    /**
     * force image generation, even when image exists
     *
     * @var bool
     */ 
		private $_forceImageGeneration = false;
		
    /**
     * set debug flag
     *
     * @var bool
     */ 
		private $_debug = false;


/*                        *
 * SOURCE IMAGE PROPERTIES *
 *                        */

    /**
     * idupl - just in use when object is inilized by an idupl (used in table con_upl)
     *
     * @var int
     */
		private $_idupl = 0;
		
    /**
     * the source image's basename
     *
     * @var string
     */
		private $_filename = '';

    /**
     * the source image's dirname - beginning AFTER upload/
     *
     * @var string
     */
		private $_dirname = '';
		
    /**
     * File extension in case we generate an image by keeping the extension. Generated by mime type.
     *
     * @var string
     */
		private $_ext = '';
		
    /**
     * source image path beginning usually with 'upload/'
     *
     * @var string
     */
		private $_src = '';
		
    /**
     * source image server path
     *
     * @var string
     */
		private $_serverPath = '';
		
    /**
     * a hash generated by parameters of the source image
     *
     * @var string
     */
		private $_hash = '';
		
    /**
     * the width of the source image. May be changed by preprocessing
     *
     * @var int
     */
		private $_width = 0;
		
    /**
     * the height of the source image. May be changed by preprocessing
     *
     * @var int
     */
		private $_height = 0;
		
    /**
     * the source mime type. i.e. Looks like a file extension, but it's taken by the mine type: 'jpg', 'gif', 'png' ...
     *
     * @var string
     */
		private $_mime = '';
		
    /**
     * width / height of the source image
     *
     * @var float
     */
		private $_ratio = 1;
		
    /**
     * last change at source image. Necessary for generating the hash
     *
     * @var timestamp
     */
		private $_filetime = 0;
		
    /**
     * size of source image. Necessary for generating the hash
     *
     * @var int
     */ 
		private $_filesize = 0;
		
    /**
     * May the source image have transparent pixels?
     *
     * @var boolean
     */ 
		private $_hasTransparency = false;
		
    /**
     * width to preprocess source image
     *
     * @var int
     */ 
		private $_preprocessWidth = 0;
		
    /**
     * file exists and is readable?
     *
     * @var bool
     */ 
		private $_exists = false;

    /**
     * height to preprocess source image
     *
     * @var int
     */ 
		private $_preprocessHeight = 0;
		
    /**
     * resource of source image
     *
     * @var resource
     */ 
		private $_img = 0;

		
/*                   *
 * OUTPUT PROPERTIES *
 *                   */
    /**
     * quality for jpg images
		 * 0 .. 100, 100 for best quality
     *
     * @var int
     */ 
		private $_jpgQuality = 80;
		
    /**
     * quality for webp images
		 * 0 .. 100, 100 for best quality
     *
     * @var int
     */ 
		private $_webpQuality = 80;
		
    /**
     * quality for png images
		 * -1 .. 9
		 * 0 for best quality, 9 for best compression (and worst quality), -1: use default (i.e. 6)
     *
     * @var int
     */ 
		private $_pngQuality = -1;
		
    /**
     * color to use for backgrounds for transparent input files and not-transparent outputfiles
     *
     * @var array of int
     */ 
		private $_backgroundColor = array('r' => 255, 'g' => 255, 'b' => 255);
		
    /**
     * foreground color - only for wbmp images
     *
     * @var array of int
     */ 
		private $_foregroundColor = array('r' => 0, 'g' => 0, 'b' => 0);
		
    /**
     * only relevant for transparent gif and png images
		 * is $trans true we get transparent pngs and gifs.
		 * is $trans false to get opac images that use backgroundcolor
		 * switch with method allowTransparency
     *
     * @var boolean
     */ 
		private $_allowTransparency = true;

    /**
     * May this class enlarge images? Switch with method allowEnlargement
     *
     * @var boolean
     */ 
		private $_allowEnlargement = false;			

    /**
     * Constructor to create an instance of this class.
     *
     * @param int or string $img
     *         idupl or source image path
     * @param int $_preprocessWidth
     *         min image width to reduce image size before resampling
     * @param int $_preprocessHeight
     *         min image hight to reduce image size before resampling
		 *         set just width OR height
		 *         set both for a min width and a min height
     */
		public function __construct($img, $_preprocessWidth = 0, $_preprocessHeight = 0) {
			
			$this->_preprocessWidth = $_preprocessWidth;
			$this->_preprocessHeight = $_preprocessHeight;

			$this->cfg = cRegistry::getConfig();
			$this->_idclient = (int) cRegistry::getClientId();
			$this->cfgClient = cRegistry::getClientConfig($this->_idclient);
			
			$this->cachePath = $this->cfgClient["cache"]["path"] . $this->relativeCacheDir;
			
			if(!is_dir($this->cachePath)) {
				mkdir($this->cachePath, 0777, true);
			}
			
			$this->cacheFrontendPath = $this->cfgClient["cache"]["frontendpath"] . $this->relativeCacheDir;
			
			
			
			$this->_exists = $this->_getImgData($img);
		}
		
    /**
     * Method to read source image information
     *
     * @param int or string $img
     *         idupl or source image path
		 * @return boolean
		 *         true if image exists and is readable, false if not
     */		
		private function _getImgData($img){
			$time_start = microtime(true);
			// empty path or without filename?
			if (empty($img) || in_array(substr($img, strlen($img)-1, 1), array('\\', '/'))) {
				$this->_exists = false;
				return false;
			}
			
			// initialsation by idupl
			if (is_numeric($img)) {
				$sql = "SELECT * FROM " . $this->cfg['tab']['upl'] . " WHERE idupl='" . (int) $img . "' AND idclient = '" . $this->_idclient . "'"; 
				$db = new cDb;
				$db->query($sql);
				if ($db->next_record()) {
					$this->_dirname = $db->f('dirname');
					$this->_filename = $db->f('filename');
					$this->_filesize = (int) $db->f('size');
				} else {
					error_log ('Can\'t find image with idupl ' . $img . ' in file ' . __FILE__ . ' line ' . __LINE__);
					return false;
				}
			} else {
			// initialisation by pathname
			// path format should fit to database entry. So we remove first 'upload/'
				$pathParts = explode('upload/', $img);
				if (count($pathParts) > 1) {
					$pathParts = array_slice($pathParts, 1);
					$path = implode('', $pathParts);
				} else {
					$path = $pathParts[0];
				}
				$path = ltrim($path, '/');
				$this->_filename = pathinfo($path, PATHINFO_BASENAME);
				$this->_dirname = ltrim(pathinfo($path, PATHINFO_DIRNAME), '.') . '/';
			}
			$src = $this->_dirname . $this->_filename;
			$src = str_replace('./', '', $src);
			if (empty($src)) {
				error_log ('Empty path! Can\'t load this image in file ' . __FILE__ . ' line ' . __LINE__);
				return false;
			}
			
			$this->_src = $this->cfgClient["upl"]["frontendpath"] . $src;
			$this->_serverPath = $this->cfgClient["upl"]["path"] . $src;
			if (!file_exists($this->_serverPath)) {
				error_log ('Can\'t find file ' . $this->_serverPath . ' in file system in file ' . __FILE__ . ' line ' . __LINE__  . ' idart:' . cRegistry::getArticleId());
				return false;
			}

			// get file properties
			$imgSize = getimagesize($this->_serverPath);
			list($this->_width, $this->_height) = $imgSize;
			if ($this->_width == 0 || $this->_height == 0) {
				error_log ('Image size is 0 (' . $this->_serverPath . ') in file ' . __FILE__ . ' line ' . __LINE__);
				return false;
			}
			
			// image ration to detect if we hve a landscape or a portrait
			$this->_ratio = $this->_width / $this->_height;
			
			// source mime type
			$this->_mime = $imgSize['mime'];
			
			$this->_filetime = filectime($this->_serverPath);
			
			if ($this->_filesize == 0) {
				$this->_filesize = filesize($this->_serverPath);
			}
			// generate source image hash using the properties above
			$this->_hash = md5($this->_src . '_' . $this->_filetime . '_' . $this->_filesize . '_' . $this->_width . '_' . $this->_height);

			$time_end = microtime(true);
			$this->log[] = ($time_end - $time_start) . ' sec.: Getting image inormation';

			return true;
		}
		
    /**
     * Method to read source image and turn it to an php image resource
     *
     *
     * @return resource
     */		
		private function _readImage() {
			$time_start = microtime(true);
			switch($this->_mime) {
				case 'image/bmp':
					if (function_exists('imagecreatefrombmp')) {
						try {						
							$this->_img = imagecreatefrombmp($this->_serverPath);
							$this->_ext = 'bmp';
						} catch (Exception $e) {
							error_log($e->getMessage() . ' in ' . __FILE__ . ' ' . __LINE__);
						}
					} else {
						$this->_detectProblem('bmp', __FILE__, __LINE__, true);
					}
					break;
				case 'image/gif':
					if (function_exists('imagecreatefromgif')) {
						try {						
							$this->_img = imagecreatefromgif($this->_serverPath);
							$this->_hasTransparency = true;
							$this->_ext = 'gif';
						} catch (Exception $e) {
							error_log($e->getMessage() . ' in ' . __FILE__ . ' ' . __LINE__);
						}
					} else {
						$this->_detectProblem('gif', __FILE__, __LINE__, true);
					}
					break;
				case 'image/jpeg':
					if (function_exists('imagecreatefromjpeg')) {
						try {
if ($this->_debug) {
	error_log('Read ' . $this->_filename . ' ' . $this->_filesize . ' idart:' . cRegistry::getArticleId());
}
							$this->_img = imagecreatefromjpeg($this->_serverPath);
							$this->_ext = 'jpg';
						} catch (Exception $e) {
							error_log($e->getMessage() . ' in ' . __FILE__ . ' ' . __LINE__);
						}
					} else {
						$this->_detectProblem('jpg', __FILE__, __LINE__, true);
					}
					break;
				case 'image/png':
					if (function_exists('imagecreatefrompng')) {
						try {						
							$this->_img = imagecreatefrompng($this->_serverPath);
							// does the source allow alpha pixels?
							// https://www.php.net/manual/en/function.getimagesize.php#105033
							$colorType = ord(@file_get_contents($this->_serverPath, NULL, NULL, 25, 1));
							$this->_hasTransparency = $colorType == 4 || $colorType == 6;
							$this->_ext = 'png';
						} catch (Exception $e) {
							error_log($e->getMessage() . ' in ' . __FILE__ . ' ' . __LINE__);
						}
					} else {
						$this->_detectProblem('png', __FILE__, __LINE__, true);
					}
					break;
				case 'image/vnd.wap.wbmp':
					if (function_exists('imagecreatefromwbmp')) {
						try {						
							$this->_img = imagecreatefromwbmp($this->_serverPath);
							$this->_ext = 'wbmp';
						} catch (Exception $e) {
							error_log($e->getMessage() . ' in ' . __FILE__ . ' ' . __LINE__);
						}
					} else {
						$this->_detectProblem('wbmp', __FILE__, __LINE__, true);
					}
					break;
				case 'image/webp':
					if (function_exists('imagecreatefrombmp')) {
						try {
							$this->_hasTransparency = true;						
							$this->_img = imagecreatefromwebp($this->_serverPath);
							$this->_ext = 'webp';
						} catch (Exception $e) {
							error_log($e->getMessage() . ' in ' . __FILE__ . ' ' . __LINE__);
						}
					} else {
						$this->_detectProblem('webp', __FILE__, __LINE__, true);
					}
					break;
				case 'image/x-xbitmap':
				case 'image/x-xbm':
					if (function_exists('imagecreatefromxbm')) {
						try {						
							$this->_img = imagecreatefromxbm($this->_serverPath);
							$this->_ext = 'xbm';
						} catch (Exception $e) {
							error_log($e->getMessage() . ' in ' . __FILE__ . ' ' . __LINE__);
						}
					} else {
						$this->_detectProblem('xbm', __FILE__, __LINE__, true);
					}
					break;
				case 'image/x-xpixmap':
				case 'image/x-xpm':
					if (function_exists('imagecreatefromxpm')) {
						try {						
							$this->_img = imagecreatefromxpm($this->_serverPath);
							$this->_ext = 'xpm';
						} catch (Exception $e) {
							error_log($e->getMessage() . ' in ' . __FILE__ . ' ' . __LINE__);
						}
					} else {
						$this->_detectProblem('xpm', __FILE__, __LINE__, true);
					}
					break;
				default: 
					error_log('File format ' . $this->_mime . ' is not supported by this script in file ' . __FILE__ . ' line ' . __LINE__);
					return false;
			}
			$this->log[] = (microtime(true) - $time_start) . ' sec.: Read Image (Size: ' . $this->_width . ' × ' . $this->_height . ')';
			if (!($this->_img)) {
				error_log('Can\t read file ' . $this->_serverPath .' or file format ' . $this->_mime . ' is not supported by this server in file ' . __FILE__ . ' line ' . __LINE__);
				return false;
			}
			// preprocess image so we get small copies much faster
			if ($this->_preprocessWidth > 0 || $this->_preprocessHeight > 0) {
				$time_start = microtime(true);
				$this->_img = $this->_getImage($this->_preprocessWidth, $this->_preprocessHeight, 'cover', $this->_hasTransparency);
				$this->_width = imagesx($this->_img);
				$this->_height = imagesy($this->_img);
				$this->log[] = (microtime(true) - $time_start) . ' sec.: Preprocessing (Size: ' . $this->_width . ' × ' . $this->_height . ')';
			}
			
			return $this->_img;
		}
		
		
    /**
     * Method to generate a hash filename for a cache file
     *
     * @param int $width
     *         dest img image width
     *
     * @param int $height
     *         dest img image height
     *
     * @param string $modus
     *         image generation modus
     *
     * @param string $ext
     *         dest img file extension
     *
     * @param string $add
     *         additional format specific properties
     *
     * @return string
     *         file specific filename 
     */		
		private function _getFilename($width, $height, $modus, $ext, $add = '') {
			$input = $width . '_' . $height . '_' . $modus . '_' . $this->_hash . '_' . $add;
			return md5($input) . '.' . $ext;
		}
		
    /**
     * Method to generate the resized image
     *
     * @param int $width
     *         dest img image width
     *
     * @param int $height
     *         dest img image height
     *
     * @param string $modus
     *         image generation modus
     *
     * @param boolean $allowAlpha
     *         does dest image format allows transparence?
     *
     * @return image resource
     *         
     */		
		private function _getImage($width, $height, $modus, $allowAlpha) {
			if (!$this->_exists) {
				return false;
			}
			if ($this->_img === 0) {
				$this->_readImage();
			}
			$time_start = microtime(true);
			// if one dimension is unset, the new image will have the same ratio like the original image
			if ($width == 0 || $height == 0) {
				if ($width > 0) {
					$cHeight = $width / $this->_ratio;
					$cWidth = $width;
				} elseif ($height > 0) {
					$cWidth = $height * $this->_ratio;
					$cHeight = $height;
				} else {
					// if both dimensions are unset the new image has the same size like the original image
					$cWidth = $this->_width;
					$cHeight = $this->_height;
				}
			} else {
				// if both sizes are given the final dimensons depends in the modus.
				// With 'cover' and 'contain' we keep the ratio and change (if neccessary) one dimension.
				// With 'crop' we create an image with the given dimensions.
				$ratio = $width / $height;
				$modus = strtolower(trim($modus));
				
				// calculate canvas (output) size $cWidth and $cHeight
				switch ($modus) {
					case 'crop':
						$cWidth = $width;
						$cHeight = $height;
						break;
					case 'contain':
						if ($ratio > $this->_ratio) {
							$cHeight = $height;
							$cWidth = $height * $this->_ratio;
						} else {
							$cWidth = $width;
							$cHeight = $width / $this->_ratio;
						}
						break;
					case 'cover':
						if ($ratio < $this->_ratio) {
							$cHeight = $height;
							$cWidth = $height * $this->_ratio;
						} else {
							$cWidth = $width;
							$cHeight = $width / $this->_ratio;
						}
						break;
					default:
						error_log('Modus \'' . $modus . '\' not defined in file ' . __FILE__ . ' line ' . __LINE__);
						return false;
				}
			}
			// output image size
			$cWidth = (int) round($cWidth);
			$cHeight = (int) round($cHeight);

			// now we draw the new image
			// calculate the offset ($srcLeft and $srcTop) and the size of the part we read($srcWidth and $srcHeight) in the source image
			switch ($modus) {
				case 'contain':
				case 'cover':
					$srcWidth = $this->_width;
					$srcHeight = $this->_height;
					$srcLeft = 0;
					$srcTop = 0;
					break;
				case 'crop':
					// wider
					if ($this->_ratio > $ratio) {
						$srcWidth = $cWidth * $this->_height / $cHeight;
						$srcLeft = ($this->_width - $srcWidth) / 2;
						$srcTop = 0;
						$srcHeight = $this->_height;
					} else {
						$srcWidth = $this->_width;
						$srcHeight = $cHeight * $this->_width / $cWidth;
						$srcLeft = 0;//$this->_width - $srcWidth;
						$srcTop = ($this->_height - $srcHeight) / 2;
					}
			}
//var_dump($srcLeft, $srcTop, $cWidth, $cHeight);

			if (($cWidth > $this->_width || $cHeight > $this->_height) && !$this->_allowEnlargement) {
				if ($modus == 'crop') {
					if ($ratio > $this->_ratio) {
						$cHeight = $cHeight * $this->_width / $cWidth;
						$cWidth  = $this->_width;
					} else {
						$cWidth  = $cWidth * $this->_height / $cHeight;
						$cHeight = $this->_height;
					}
				} else {
					$cWidth  = $this->_width;
					$cHeight = $this->_height;
				}
//var_dump($srcLeft, $srcTop, $cWidth, $cHeight);
			}
			// turn all values to integers
			$srcLeft   = (int) round($srcLeft);
			$srcTop    = (int) round($srcTop);
			$srcWidth  = (int) round($srcWidth);
			$srcHeight = (int) round($srcHeight);
			$cWidth    = (int) round($cWidth);
			$cHeight   = (int) round($cHeight);

			// generate dest image
			$im = imagecreatetruecolor($cWidth, $cHeight);

			// fill background if neccessary (source allows alpha, output doesn't)
			if (!($this->_allowTransparency && $allowAlpha) && $this->_hasTransparency) {
				$bg = imagecolorallocate($im, $this->_backgroundColor['r'], $this->_backgroundColor['g'], $this->_backgroundColor['b']);
				imagefilledrectangle($im, 0, 0, $cWidth, $cHeight, $bg);
			} elseif ($allowAlpha && $this->_hasTransparency) {
				// turn transparent image to a transparent image
				imagecolortransparent($im, imagecolorallocatealpha($im, 0, 0, 0, 127));
				imagealphablending($im, false);
				imagesavealpha($im, true);				
			}
			//now the magic happens. We generate the new image.
			imagecopyresampled($im, $this->_img, 0, 0, $srcLeft, $srcTop, $cWidth, $cHeight, $srcWidth, $srcHeight);
			$time_end = microtime(true);
			$this->log[] = ($time_end - $time_start) . ' sec.: Draw image. Size: ' . $cWidth . ' × ' .  $cHeight;
							
			return $im;
		}
		
    /**
     * write error messages
     *
     * @param int $ext
     *         extension / type of dest image
     *
     * @return none
     *         
     */		
		private function _detectProblem($ext, $file = '', $line = '', $read = false) {
			if ($this->_exists) {
				if ($this->typeAvailable($ext)) {
					error_log('An error occured while trying to ' . ($read ? 'read' : 'generate') . ' a ' . $ext . ' file' . ($file ? ' in file ' . $file . ($line ? ' line ' . $line : '') : ''));
				} else {
					error_log('Server doesn\'t support ' . $ext . ' files' . ($file ? ' in file ' . $file . ($line ? ' line ' . $line : '') : ''));
				}
			}
		}
		
    /**
     * test if a filetype is available
     *
     * @param int $ext
     *         extension / type of dest image
     *
     * @return boolean
     *         
     */
		public function typeAvailable($ext) {
			$ext = strtolower($ext);
			switch ($ext) {
				case 'bmp':  return (bool) imagetypes() & IMG_BMP;
				case 'gif':  return (bool) imagetypes() & IMG_GIF;
				case 'jpg':
				case 'jpeg': return (bool) imagetypes() & IMG_JPG;
				case 'png':  return (bool) imagetypes() & IMG_PNG;
				case 'wbmp': return (bool) imagetypes() & IMG_WBMP;
				case 'xpm':  return (bool) imagetypes() & IMG_XPM;
				case 'webp': return (bool) imagetypes() & IMG_WEBP;
				default: return false;
			}
		}

    /**
     * method getBmp returns the path to a cached bmp image in the given size.
     *
     * @param int $width
     *         output width
     *
     * @param int $height
     *         output height
     *
     * @param string $modus
     *         'cover', 'contain' or 'crop'
     *
     * @return string
     *         	path to cached file
     */
		public function getBmp($width = 0, $height = 0, $modus = 'cover') {
			// collect parameters to build cache name.
			$add = '';
			if ($this->_hasTransparency) {
				$add = implode('_', $this->_backgroundColor);
			}

			// generate image if necessary
			$newFileName = $this->_getFilename($width, $height, $modus, 'bmp', $add);
			
			if (!file_exists($this->cachePath . $newFileName)) {
				$new = $this->_getImage($width, $height, $modus, false);
				
				if ($new && function_exists('imagebmp') && imagebmp($new, $this->cachePath . $newFileName, true)) {
					imagedestroy($new);
					return $this->cacheFrontendPath . $newFileName;
				} else {
					$this->_detectProblem('bmp', __FILE__, __LINE__);
					return false;
				}
			} else {
				return $this->cacheFrontendPath . $newFileName;
			}
		}

    /**
     * method returns the path to a cached gif image in the given size.
     *
     * @param int $width
     *         output width
     *
     * @param int $height
     *         output height
     *
     * @param string $modus
     *         'cover', 'contain' or 'crop'
     *
     * @return string
     *         	path to cached file
     */
		public function getGif($width = 0, $height = 0, $modus = 'cover') {
			// collect parameters to build cache name.
			$add = $this->_allowTransparency ? '_allowTransparency' : '';
			if ($this->_hasTransparency && $this->_allowTransparency) {
				$add = '_' . implode('_', $this->_backgroundColor);
			}

			// generate image if necessary
			$newFileName = $this->_getFilename($width, $height, $modus, 'gif', $add);
			
			if (!file_exists($this->cachePath . $newFileName) || $this->_forceImageGeneration) {
				$new = $this->_getImage($width, $height, $modus, true);
				
				if ($new && function_exists('imagegif') && imagegif($new, $this->cachePath . $newFileName)) {
					imagedestroy($new);
					return $this->cacheFrontendPath . $newFileName;
				} else {
					$this->_detectProblem('gif', __FILE__, __LINE__);
					return false;
				}
			} else {
				return $this->cacheFrontendPath . $newFileName;
			}
		}

    /**
     * method returns the path to a cached jpg image in the given size.
     *
     * @param int $width
     *         output width
     *
     * @param int $height
     *         output height
     *
     * @param string $modus
     *         'cover', 'contain' or 'crop'
     *
     * @return string
     *         	path to cached file
     */
		public function getJpg($width = 0, $height = 0, $modus = 'cover') {
			// generate new filename using all parameters
			$add = $this->_jpgQuality;
			if ($this->_hasTransparency) {
				$add .= '_' . implode('_', $this->_backgroundColor);
			}

			// generate image if necessary
			$newFileName = $this->_getFilename($width, $height, $modus, 'jpg', $add);

			if (!file_exists($this->cachePath . $newFileName) || $this->_forceImageGeneration) {
				$new = $this->_getImage($width, $height, $modus, false);
if ($this->_debug) {
	error_log($this->_filename . ' ' . $this->_filesize . ' JPG' . ' idart:' . cRegistry::getArticleId());
}
				if ($new && function_exists('imagejpeg') && imagejpeg($new, $this->cachePath . $newFileName, $this->_jpgQuality)) {
					imagedestroy($new);
					return $this->cacheFrontendPath . $newFileName;
				} else {
					$this->_detectProblem('jpg', __FILE__, __LINE__);
					return false;
				}
			} else {
				return $this->cacheFrontendPath . $newFileName;
			}
		}
		
    /**
     * alias method for getJpg - returns the path to a cached jpg image in the given size.
     *
     * @param int $width
     *         output width
     *
     * @param int $height
     *         output height
     *
     * @param string $modus
     *         'cover', 'contain' or 'crop'
     *
     * @return string
     *         	path to cached file
     */
		public function getJpeg($width = 0, $height = 0, $modus = 'cover') {
			return $this->getJpg($width, $height, $modus);
		}

    /**
     * method returns the path to a cached png image in the given size.
     *
     * @param int $width
     *         output width
     *
     * @param int $height
     *         output height
     *
     * @param string $modus
     *         'cover', 'contain' or 'crop'
     *
     * @return string
     *         	path to cached file
,     */
		public function getPng($width = 0, $height = 0, $modus = 'cover') {
			// collect parameters to build cache name.
			$add = $this->_pngQuality . ($this->_allowTransparency ? '__allowTransparency' : '');
			if ($this->_hasTransparency && !$this->_allowTransparency) {
				$add .= '_' . implode('_', $this->_backgroundColor);
			}
			$newFileName = $this->_getFilename($width, $height, $modus, 'png', $add);
			
			// generate image if necessary
			if (!file_exists($this->cachePath . $newFileName) || $this->_forceImageGeneration) {
				$new = $this->_getImage($width, $height, $modus, true);
				
				if ($new && function_exists('imagepng') && imagepng($new, $this->cachePath . $newFileName, $this->_pngQuality)) {
					imagedestroy($new);
					return $this->cacheFrontendPath . $newFileName;
				} else {
					$this->_detectProblem('png', __FILE__, __LINE__);
					return false;
				}
			} else {
				return $this->cacheFrontendPath . $newFileName;
			}
		}		

    /**
     * method returns the path to a cached wbmp image in the given size.
		 * USUAL BROWSERS DON'T SUPPORT THIS FORMAT!
     *
     * @param int $width
     *         output width
     *
     * @param int $height
     *         output height
     *
     * @param string $modus
     *         'cover', 'contain' or 'crop'
     *
     * @return string
     *         	path to cached file
     */
		public function getWbmp($width = 0, $height = 0, $modus = 'cover') {
			// collect parameters to build cache name.
			$add = '';
			if ($this->_hasTransparency) {
				$add .= '_' . implode('_', $this->_foregroundColor);
			}

			// generate image if necessary
			$newFileName = $this->_getFilename($width, $height, $modus, 'wbmp', $add);
			
			if (!file_exists($this->cachePath . $newFileName) || $this->_forceImageGeneration) {
				$new = $this->_getImage($width, $height, $modus, false);
				$foreground_color = imagecolorallocate($new, $this->_foregroundColor['r'], $this->_foregroundColor['g'], $this->_foregroundColor['b']);
				if ($new && function_exists('imagewbmp') && imagewbmp($new, $this->cachePath . $newFileName, $foreground_color)) {
					imagedestroy($new);
					return $this->cacheFrontendPath . $newFileName;
				} else {
					$this->_detectProblem('wbmp', __FILE__, __LINE__);
					return false;
				}
			} else {
				return $this->cacheFrontendPath . $newFileName;
			}
		}
		
    /**
     * method returns the path to a cached Xbm image in the given size.
		 * USUAL BROWSERS DON'T SUPPORT THIS FORMAT!
     *
     * @param int $width
     *         output width
     *
     * @param int $height
     *         output height
     *
     * @param string $modus
     *         'cover', 'contain' or 'crop'
     *
     * @return string
     *         	path to cached file
     */
		public function getXbm($width = 0, $height = 0, $modus = 'cover') {
			// collect parameters to build cache name.
			$add = '';
			if ($this->_hasTransparency) {
				$add .= '_' . implode('_', $this->_foregroundColor);
			}

			// generate image if necessary
			$newFileName = $this->_getFilename($width, $height, $modus, 'xbm', $add);
			
			if (!file_exists($this->cachePath . $newFileName) || $this->_forceImageGeneration) {
				$new = $this->_getImage($width, $height, $modus, false);
				$foreground_color = imagecolorallocate($new, $this->_foregroundColor['r'], $this->_foregroundColor['g'], $this->_foregroundColor['b']);
				if ($new && function_exists('imagexbm') && imagexbm($new, $this->cachePath . $newFileName, $foreground_color)) {
					imagedestroy($new);
					return $this->cacheFrontendPath . $newFileName;
				} else {
					$this->_detectProblem('xbm', __FILE__, __LINE__);
					return false;
				}
			} else {
				return $this->cacheFrontendPath . $newFileName;
			}
		}
		
    /**
     * method returns the path to a cached webp image in the given size.
     *
     * @param int $width
     *         output width
     *
     * @param int $height
     *         output height
     *
     * @param string $modus
     *         'cover', 'contain' or 'crop'
     *
     * @return string
     *         	path to cached file
     */
		public function getWebp($width = 0, $height = 0, $modus = 'cover') {
			// collect parameters to build cache name.
			$add = $this->_webpQuality . ($this->_allowTransparency ? '__allowTransparency' : '');
			if ($this->_hasTransparency) {
				$add .= '_' . implode('_', $this->_backgroundColor);
			}

			// generate image if necessary
			$newFileName = $this->_getFilename($width, $height, $modus, 'webp', $add);
			
			if (!file_exists($this->cachePath . $newFileName) || $this->_forceImageGeneration) {
				$new = $this->_getImage($width, $height, $modus, true);
if ($this->_debug) {
	error_log($this->_filename . ' ' . $this->_filesize . ' WEBP' . ' idart:' . cRegistry::getArticleId());
	if ($this->_filename == 'slider05.jpg') {
		error_log($this->_filename . ' ' . $this->_filesize . ' WEBP width:' . $width . ' height: ' .  $height . ' name: ' . $newFileName);
	}
}
				if ($new && function_exists('imagewebp') && imagewebp($new, $this->cachePath . $newFileName, $this->_webpQuality)) {
					imagedestroy($new);
					return $this->cacheFrontendPath . $newFileName;
				} else {
					$this->_detectProblem('webp', __FILE__, __LINE__);
					return false;
				}
			} else {
				return $this->cacheFrontendPath . $newFileName;
			}
		}

    /**
     * method returns the path to a cached image in the given size in file format of a given mime type  or in the source files type
     *
     * @param int $width
     *         output width
     *
     * @param int $height
     *         output height
     *
     * @param string $modus
     *         'cover', 'contain' or 'crop'
     *
     * @param string $type
     *         mimetype, leave empty to take the type of the original image
     *
     * @return string
     *         	path to cached file
     */
		public function getImage($width = 0, $height = 0, $modus = 'cover', $type = '') {
			$type = empty($type) ? $this->_mime : $type;
			$type = strtolower($type);
			if (strpos($type, 'image/') === false) {
				$type = 'image/' . $type;
			}
			switch($type) {
				case 'image/bmp':
					return $this->getBmp($width, $height, $modus);
				case 'image/gif':
					return $this->getGif($width, $height, $modus);
				case 'image/jpeg':
					return $this->getJpg($width, $height, $modus);
				case 'image/png':
					return $this->getPng($width, $height, $modus);
				case 'image/vnd.wap.wbmp':
					return $this->getWbmp($width, $height, $modus);
				case 'image/webp':
					return $this->getWebp($width, $height, $modus);
				case 'image/x-xbitmap':
				case 'image/x-xbm':
				case 'image/x-xpixmap':
				case 'image/x-xpm':
					return $this->getXbm($width, $height, $modus);
				default:
					return $this->getJpg($width, $height, $modus);
			}
		}		

		
    /**
     * method to set the quality for jpg files.
     *
     * @param int $jpgQuality
     *         jpg quality (0..100)
     *
     * @return $this
     */
		public function setJpgQuality ($jpgQuality = 80) {
			$this->_jpgQuality = (int) round(max(min(100, $jpgQuality), 0));
			return $this;
		}

    /**
     * method to set the quality for jpg files.
     *
     * @param int $jpgQuality
     *         jpg quality (0..100)
     *
     * @return $this
     */
		public function setJpegQuality ($jpegQuality = 80) {
			$this->setJpgQuality($jpegQuality);
			return $this;
		}

    /**
     * method to set the quality for png files.
		 * set value -1 .. 9. Integers only!
		 * 0 for best quality, 9 for best compression (and worst quality), -1: use default (i.e. 6)
     *
     * @param int $pngQuality
     *         png quality (-1..9)
     *
     * @return $this
     */
		public function setPngQuality ($pngQuality = -1) {
			$this->_pngQuality = (int) round(max(min(9, $pngQuality), -1));
			return $this;
		}
		
    /**
     * method to set the quality for webp files.
     *
     * @param int $webpQuality
     *         webp quality (0..100)
     *
     * @return $this
     */
		public function setWebpQuality ($webpQuality = 80) {
			$this->_webpQuality = (int) round(max(min(100, $webpQuality), 0));
			return $this;
		}

    /**
     * method to set backgroundcolor. In use when we turn a transparent source to a not transparent cache file.
		 * set RGB data 0..255
		 *
     * @param int $r
     *         red channel, 0..255
		 *
     * @param int $g
     *         green channel, 0..255
		 *
     * @param int $b
     *         blue channel, 0..255
     *
     * @return $this
     */
		public function setBackgroundColor($r, $g, $b) {
			// set 3 values 0 .. 255 for background color. Integers only.
			$this->_backgroundColor['r'] = (int) round(max(min(255, $r), 0));
			$this->_backgroundColor['g'] = (int) round(max(min(255, $g), 0));
			$this->_backgroundColor['b'] = (int) round(max(min(255, $b), 0));
			return $this;
		}
		
    /**
     * method to set foregroundcolor. In use when you generate a wbmp, xbm or xpm file.
		 * set RGB data 0..255
		 *
     * @param int $r
     *         red channel, 0..255
		 *
     * @param int $g
     *         green channel, 0..255
		 *
     * @param int $b
     *         blue channel, 0..255
     *
     * @return $this
     */
		public function setForegroundColor($r, $g, $b) {
			// set 3 values 0 .. 255 for background color. Integers only.
			$this->_foregroundColor['r'] = (int) round(max(min(255, $r), 0));
			$this->_foregroundColor['g'] = (int) round(max(min(255, $g), 0));
			$this->_foregroundColor['b'] = (int) round(max(min(255, $b), 0));
			return $this;
		}
		
    /**
     * method allowTransparency is relevant for tarnsparent gif and png images. 
		 * set $trans true to get transparent output images.
		 * set $trans false to get opac images that use backgroundcolor
		 *
     * @param boolean $trans
     *
     * @return $this
     */
		public function allowTransparency($trans = true) {
			$this->_allowTransparency = (bool) $trans;
			return $this;
		}
		
		/**
			* method getMetaData lists meta data from con_upl_meta
			* set $idlang if you don't want data in current language
			*
			* @param int $idlang 
			*
			* @return ass. array
		 	*/
			
		public function getMetaData($idlang = 0) {
			$db = new cDb;
			if ($idlang == 0) {
					$idlang = cRegistry::getLanguageId();
			}
			$sql = "
				SELECT
					cum.*
				FROM " . $this->cfg["tab"]["upl"] . " as cu
				LEFT JOIN " . $this->cfg["tab"]["upl_meta"] . " as cum ON
					cum.idupl = cu.idupl
				AND
					cum.idlang = '" . (int) $idlang . "'
				WHERE 
					cu.dirname = '" . $db->escape($this->_dirname) . "'
				AND
					cu.filename = '" . $db->escape($this->_filename) . "'";

			$out = array();
			if ($db->query($sql)) {
				$db->next_record();
				$out['id_upl_meta'] = $db->f('id_upl_meta');
				$out['idupl'] = $db->f('idupl');
				$out['idlang'] = $db->f('idlang');
				$out['medianame'] = $db->f('medianame');
				$out['description'] = $db->f('description');
				$out['keywords'] = $db->f('keywords');
				$out['internal_notice'] = $db->f('internal_notice');
				$out['author'] = $db->f('author');
				$out['created'] = $db->f('created');
				$out['modified'] = $db->f('modified');
				$out['modifiedby'] = $db->f('modifiedby');
				$out['copyright'] = $db->f('copyright');
			}
			return $out;
		}

    /**
     * method returns the source image path beginning usually with 'upload/'
		 *
     * @return string
     */
		public function getSrc() {
			return $this->_src;
		}
		
    /**
     * true if image exists and is readable, false if not
		 *
		 * @return boolean
		 */
		public function exists() {
			return $this->_exists;
		}

    /**
     * sets the permissen to enlarge small source images
		 *
		 * @param boolean $allow 
		 *
		 * @return $this
		 */
		public function allowEnlargement($allow = true) {
			$this->_allowEnlargement = (bool) $allow;
			return $this; 
		}

    /**
     * destroys source image ressource.
		 *
		 * @return null
		 */
		public function free() {
			if ($this->_img){
				try {
					imagedestroy($this->_img);
				} catch(Exception $e) {
				}
			}
		}

    /**
     * method returns the log entries in a list structure
		 *
     * @return string
     */
		public function getLog() {
			if (count($this->log) == 0) {
				return 'No entries';
			}
			return '<ul><li>' . implode('</li><li>', $this->log) . '</li></ul>';
		}
	}
?>